@@ -1,6 +1,8 @@
 
 import datetime
+import logging
 import random
+from traceback import format_exception, format_tb
 from typing import List,Callable
 import pytz
 import requests
@@ -8,7 +10,7 @@ import hashlib
 import sys
 import asyncio
 from dataclasses import dataclass
-from threading import Lock, Thread
+from threading import Lock, Thread,Event
 from PIL import Image
 import numpy as np
 from queue import SimpleQueue
@@ -16,7 +18,7 @@ import io
 from contextlib import contextmanager
 import time
 from hashlib import sha256
-
+from secrets import token_hex
 
 
 
@@ -36,7 +38,7 @@ def buttons_transform(buttons):
         for column in row:
             
             
-            column._id = sha256(f'{random.random()}'.encode('utf-8')).hexdigest()[:12]
+            column._id = token_hex(6)
             rowButtons.append({
                 "button_id":column._id,
                 "text":column.text
@@ -205,6 +207,7 @@ class PytificationsMessage:
 
 
 class PytificationsRemoteController:
+    _instance = None
     def __init__(self,name) -> None:
         pass
 
@@ -251,6 +254,7 @@ class PytificationsOptions:
     def no_update_in_server():
         """
         INTERNAL
+
         Helper decorator to avoid updating the alias of the script when inside the with statement
         """
         try:
@@ -263,19 +267,21 @@ class PytificationsOptions:
     @contextmanager
     def default_options():
         """
-        Helper decorator to keep the default options while inside the with statement
+        INTERNAL
+
+        Helper decorator to keep the default options while inside the with statement without sending alias
         """
         current_options = Pytifications._options
         
         with PytificationsOptions.no_update_in_server() as c:
             try:
-                Pytifications.set_options(PytificationsOptions())
+                Pytifications.set_options(PytificationsOptions(should_send_script_alias=False))
                 yield None
             finally:
                 Pytifications.set_options(current_options)
 
 
-    def __init__(self,send_current_time_on_message=False,send_app_run_time_on_message = False,script_alias = "",should_print_sent_messages=False) -> None:
+    def __init__(self,send_current_time_on_message=False,send_app_run_time_on_message = False,script_alias = "",should_print_sent_messages=False,should_send_script_alias=True) -> None:
         """
         Data class for the options in Pytifications
 
@@ -287,12 +293,17 @@ class PytificationsOptions:
             script_alias: (:obj:`str`) alias to use when sending the message. Any spaces will be replaced with an underscore. Will appear on the top of the messages as "Message sent from __alias_here__:"  and will be used to send commands to your script
         
             should_print_sent_messages: (:obj:`bool`) whether to print to console when a message is sent
+
+            should_send_script_alias: (:obj:`bool`) wether to send the alias of the script at the top of every message
         """
         
         self._send_app_run_time_on_message = send_app_run_time_on_message
         self._script_alias = script_alias.replace(" ","_")
+        self._script_alias = token_hex(6) if script_alias == "" else script_alias
+
         self._send_current_time_on_message = send_current_time_on_message
         self._should_print_sent_messages = should_print_sent_messages
+        self._should_send_script_alias = should_send_script_alias
         
     
 
@@ -307,7 +318,7 @@ class PytificationsOptions:
         if self._send_current_time_on_message:
             string = f'{string}\ncurrent_time:\n{(datetime.datetime.now(pytz.UTC) + Pytifications._prefered_timezone).strftime("%H:%M:%S")}'
 
-        if self._script_alias != "":
+        if self._should_send_script_alias:
             string = f'Message sent from "{self._script_alias}":\n\n{string}'
         
         return string
@@ -326,7 +337,7 @@ def send_message(request_data,photo: Image,return_data: PytificationsMessage | P
         
 
         if res.status_code != 200:
-            print(f'could not send message. reason: {res.reason}')
+            print(f'could not send message. reason: {res.text}')
             return False
 
         Pytifications._last_message_id = int(res.text)
@@ -349,7 +360,11 @@ def edit_message(request_data,return_data: PytificationsMessage | PytificationsM
         return
     with PytificationsOptions.with_custom_options(current_options):
         try:     
-            requests.patch('https://pytifications.herokuapp.com/edit_message',json=request_data)
+            res = requests.patch('https://pytifications.herokuapp.com/edit_message',json=request_data)
+
+            if res.status_code != 200:
+                print(f'could not edit message. reason: {res.text}')
+                return False
         except Exception as e:
             print(f'Found exception while editing message: {e}')
             return False
@@ -421,7 +436,10 @@ class Pytifications:
     _options = PytificationsOptions()
     _commands_mutex = Lock()
     _callbacks_mutex = Lock()
-
+    _background_thread_event = Event()
+    _background_thread = None
+    _traceback_message = ""
+    _user_id = 0
 
 
     @staticmethod
@@ -501,7 +519,8 @@ class Pytifications:
                 "username":Pytifications._login,
                 "password_hash":hashlib.sha256(Pytifications._password.encode('utf-8')).hexdigest(),
                 "process_id":Pytifications._process_id,
-                "process_name":options._script_alias
+                "process_name":options._script_alias,
+                "user_id":Pytifications._process_id
             })
         Pytifications._options = options
     
@@ -539,14 +558,17 @@ class Pytifications:
             :obj:`True`if login was successful else :obj:`False`
         """
 
+        
+
         Pytifications._logged_in = False
 
+
         try:
             res = requests.post('https://pytifications.herokuapp.com/initialize_script',json={
                 "username":login,
                 "password_hash":hashlib.sha256(password.encode('utf-8')).hexdigest(),
-                "process_name":hashlib.sha256(sys.argv[0].encode('utf-8')).hexdigest()[:10] if Pytifications._options._script_alias == "" else Pytifications._options._script_alias,
-                "process_language":'python'
+                "process_name":Pytifications._options._script_alias,
+                "process_language":'python',
             })
         except Exception as e:
             print(f'Found exception while logging in: {e}')
@@ -561,41 +583,83 @@ class Pytifications:
             json = res.json()
             Pytifications._logged_in = True
             Pytifications._process_id = json['id']
+            Pytifications._user_id = json['user_id']
             Pytifications._prefered_timezone = datetime.timedelta(hours=json['prefered_timezone'])
 
             print(f'success logging in to pytifications! script id = {Pytifications._process_id}')
 
-        Thread(target=Pytifications._check_if_any_events_called,daemon=True,args=(Pytifications._callbacks_mutex,Pytifications._commands_mutex)).start()
+        sys.excepthook = Pytifications._handle_global_exceptions
+
+        Pytifications._background_thread = Thread(target=Pytifications._check_if_any_events_called,daemon=True,args=(Pytifications._background_thread_event,Pytifications._callbacks_mutex,Pytifications._commands_mutex))
+        Pytifications._background_thread.start()
         
         return True
-
     
     @staticmethod
-    def _check_if_any_events_called(callback_lock,command_lock):
+    def _handle_global_exceptions(e_type, e_value, e_traceback):
+        logging.getLogger(__name__).critical("Unhandled exception", exc_info=(e_type, e_value, e_traceback))
+        if Pytifications.am_i_logged_in():
+            Pytifications._traceback_message = f'An error ocurred in script/application with alias "{Pytifications._options._script_alias}":\n\nError type: {e_type}\n\nError value: {e_value}\n\nTraceback: {"".join(format_tb(e_traceback))}'
+            Pytifications._background_thread_event.set()
+            while Pytifications._background_thread.is_alive():
+                pass
+    
+    @staticmethod
+    def _check_if_any_events_called(stop_event: Event,callback_lock,command_lock):
         """
         INTERNAL
         
         Used for the inside logic to check if any callbacks or commands to be called
         """
         while True:
+            if not Pytifications.am_i_logged_in():
+                continue
+            if stop_event.is_set():
+                requests.post('https://pytifications.herokuapp.com/delete_process',json={
+                    "username":Pytifications._login,
+                    "password_hash":hashlib.sha256(Pytifications._password.encode('utf-8')).hexdigest(),
+                    "process_id":Pytifications._process_id,
+                    "traceback_message":Pytifications._traceback_message,
+                    "user_id":Pytifications._user_id
+                })
+                break
+
             time.sleep(1)
             if not Pytifications.am_i_logged_in():
                 continue
+            if stop_event.is_set():
+                requests.post('https://pytifications.herokuapp.com/delete_process',json={
+                    "username":Pytifications._login,
+                    "password_hash":hashlib.sha256(Pytifications._password.encode('utf-8')).hexdigest(),
+                    "process_id":Pytifications._process_id,
+                    "traceback_message":Pytifications._traceback_message,
+                    "user_id":Pytifications._user_id
+                })
+                break
             
-            try:
-                while not Pytifications._message_pool.empty():
+            while not Pytifications._message_pool.empty():
+                try:
                     Pytifications._message_pool.get().evaluate()
-                    
-            except Exception as e:
-                print(f'Error found while updating message pool, please report to the developer! {e}')
-                pass
+                except Exception as e:
+                    print(f'Error found while updating message pool, please report to the developer! {e}')
+                    pass
 
             time.sleep(1)
+            if stop_event.is_set():
+                requests.post('https://pytifications.herokuapp.com/delete_process',json={
+                    "username":Pytifications._login,
+                    "password_hash":hashlib.sha256(Pytifications._password.encode('utf-8')).hexdigest(),
+                    "process_id":Pytifications._process_id,
+                    "traceback_message":Pytifications._traceback_message,
+                    "user_id":Pytifications._user_id
+                })
+                break
             try:
                 res = requests.get('https://pytifications.herokuapp.com/get_callbacks',json={
                     "username":Pytifications._login,
                     "password_hash":hashlib.sha256(Pytifications._password.encode('utf-8')).hexdigest(),
-                    "process_id":Pytifications._process_id
+                    "process_id":Pytifications._process_id,
+                    "user_id":Pytifications._user_id
                 })
             except Exception as e:
                 print(f'Error found while requesting to get callbacks, please contact the developer! {e}')
@@ -607,7 +671,7 @@ class Pytifications:
                     if Pytifications._synchronous:
                         #print(f'registered_callbacks = {Pytifications._registered_callbacks}')
                         #print(f'got json = {callbacks}')
-                        with Pytifications._callbacks_mutex:
+                        with callback_lock:
                             Pytifications._callbacks_to_call_synchronous.append({
                                 "function":Pytifications._registered_callbacks[item["button_id"]]["function"],
                                 "args":(Pytifications._registered_callbacks[item['button_id']]['args'] + item["args"])
@@ -618,7 +682,7 @@ class Pytifications:
                 for item in commands:
                     if item['command'] in Pytifications._registered_commands:
                         if Pytifications._synchronous:
-                                with Pytifications._commands_mutex:
+                                with command_lock:
                                     Pytifications._commands_to_call_synchronous.append({
                                         "function":Pytifications._registered_commands[item['command']]["function"],
                                         "args":item['args'],
@@ -756,4 +820,9 @@ class Pytifications:
         """
         EXPERIMENTAL: DO NOT USE
         """
-        return PytificationsRemoteController(name)
\ No newline at end of file
+
+        if PytificationsRemoteController._instance == None:
+            return PytificationsRemoteController(name)
+        else:
+            return PytificationsRemoteController._instance
+    
